//--------------------------------------------------------------------------------------
// File: FFT.fx
//
//--------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------
// Textures and Samplers
//--------------------------------------------------------------------------------------
Texture2D SourceImgTexture;
Texture2D ButterflyTexture; 

float ButterflyPassNumber;
bool GreyScale;
float ScaleFactor;

BlendState NoBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
};

DepthStencilState EnableDepth
{
    DepthEnable = TRUE;
    DepthWriteMask = ALL;
};

DepthStencilState DisableDepth
{
    DepthEnable = FALSE;
    DepthWriteMask = ZERO;
};

DepthStencilState DisableDepthWrite
{
    DepthEnable = TRUE;
    DepthWriteMask = ZERO;
};

DepthStencilState DisableDepthTest
{
    DepthEnable = TRUE;
    DepthWriteMask = ALL;
    DepthFunc = ALWAYS;
};

SamplerState g_samLinear
{
    Filter = MIN_MAG_MIP_POINT;
    AddressU = Mirror;
    AddressV = Mirror;
};

SamplerState ButterflyTextureSampler
{
    Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
};


// FFT stuff.......

struct SCR_VS_INPUT
{
    float3 Pos          : POSITION;         //position
    float2 Tex          : TEXCOORD0;        //texture coordinate
};

struct SCR_VS_OUTPUT
{
    float4 Pos : SV_POSITION;
    float2 Tex : TEXCOORD1;
};


SCR_VS_OUTPUT HButterflyVS( SCR_VS_INPUT In )
{
	SCR_VS_OUTPUT Out = (SCR_VS_OUTPUT)0;
	Out.Pos = float4(In.Pos, 1);
	Out.Tex = In.Tex;
	
	return Out;	
}

float4 HButterflyPS( SCR_VS_OUTPUT In ) : SV_Target
{
	float2 coord = In.Tex;
	// get the scrambled indices, that give the location for a and b for computing a+wb or a-wb		
	float4 IndicesAndWeights = ButterflyTexture.Sample(ButterflyTextureSampler, float2(coord.x, ButterflyPassNumber));
	float2 Indices = IndicesAndWeights.rg;
	float2 Weights = IndicesAndWeights.ba;

	float2 a1, b1;
	a1 = SourceImgTexture.Sample( g_samLinear, float2(Indices.x, coord.y) ).ra;
	b1 = SourceImgTexture.Sample( g_samLinear, float2(Indices.y, coord.y) ).ra;

	// perform a+w*b or a-w*b
	// w*b (this is complex multiplication)
	float2 res;
	
	// w*b => (wr + i*wi)(br + i*bi) => [(wr*br-wi*bi)+i(wi*br+wr*bi)]
	res.r = Weights.r*b1.r - Weights.g*b1.g;
	res.g = Weights.g*b1.r + Weights.r*b1.g;
	
    a1 = a1 + res;

	return a1.xxxy;
}

technique10 HorizontalButterfly
{
	pass P0
	{
        SetVertexShader( CompileShader( vs_4_0, HButterflyVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0,  HButterflyPS() ) );
        
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepthTest, 0 );
	}
}



SCR_VS_OUTPUT VButterflyVS( SCR_VS_INPUT In )
{
	SCR_VS_OUTPUT Out = (SCR_VS_OUTPUT)0;
	Out.Pos = float4(In.Pos, 1);
	Out.Tex = In.Tex;
	
	return Out;	
}

float4 VButterflyPS( SCR_VS_OUTPUT In ) : SV_Target
{
	float2 coord = In.Tex;

	// get the scrambled indices, that give the location for a and b for computing a+wb or a-wb		
	float4 IndicesAndWeights = ButterflyTexture.Sample(ButterflyTextureSampler, float2(coord.y, ButterflyPassNumber));
	float2 Indices = IndicesAndWeights.rg;
	float2 Weights = IndicesAndWeights.ba;

	float2 a1, b1;
	a1 = SourceImgTexture.Sample( g_samLinear, float2(coord.x, Indices.x) ).ra;
	b1 = SourceImgTexture.Sample( g_samLinear, float2(coord.x, Indices.y) ).ra;

	// perform a+w*b or a-w*b
	// w*b (this is complex multiplication)
	float2 res;
	
	// w*b => (wr + i*wi)(br + i*bi) => [(wr*br-wi*bi)+i(wi*br+wr*bi)]
	res.r = Weights.r*b1.r - Weights.g*b1.g;
	res.g = Weights.g*b1.r + Weights.r*b1.g;
	
    a1 = a1 + res;

	return a1.xxxy;
}

technique10 VerticalButterfly
{
	pass P0
	{
        SetVertexShader( CompileShader( vs_4_0, VButterflyVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0,  VButterflyPS() ) );
        
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepthTest, 0 );
	}
}


SCR_VS_OUTPUT ScaleVS( SCR_VS_INPUT In )
{
	SCR_VS_OUTPUT Out = (SCR_VS_OUTPUT)0;
	Out.Pos = float4(In.Pos, 1);
	Out.Tex = 1-In.Tex;
	
	return Out;	
}

float4 ScalePS( SCR_VS_OUTPUT In ) : SV_Target
{
	float4 Image = SourceImgTexture.Sample( g_samLinear, In.Tex );

	Image *= ScaleFactor;
	
	return Image;
}

technique10 Scale
{
	pass P0
	{
        SetVertexShader( CompileShader( vs_4_0, ScaleVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0,  ScalePS() ) );
        
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepthTest, 0 );        
	}
}


// 0.1 * log(1+sqrt(a1.r*a1.r+a1.g*a1.g))
SCR_VS_OUTPUT DisplayVS( SCR_VS_INPUT In )
{
	SCR_VS_OUTPUT Out = (SCR_VS_OUTPUT)0;
	Out.Pos = float4(In.Pos, 1);
	Out.Tex = In.Tex;
	
	return Out;	
}

float4 DisplayPS( SCR_VS_OUTPUT In ) : SV_Target
{
	float4 Image;
	if (GreyScale)
	{
		Image = SourceImgTexture.Sample( g_samLinear, In.Tex );	
	}
	else
	{
		Image = SourceImgTexture.Sample( g_samLinear, In.Tex - 0.5 );

		float t = 0.1*log2(1+sqrt(Image.r*Image.r+Image.a*Image.a));
		
		Image = float4(t, t, t, t);
	}
	return Image;
}

technique10 Display
{
	pass P0
	{
        SetVertexShader( CompileShader( vs_4_0, DisplayVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0,  DisplayPS() ) );
        
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepthTest, 0 );        
	}
}

SCR_VS_OUTPUT ToGreyScaleVS( SCR_VS_INPUT In )
{
	SCR_VS_OUTPUT Out = (SCR_VS_OUTPUT)0;
	Out.Pos = float4(In.Pos, 1);
	Out.Tex = In.Tex;
	
	return Out;	
}

float4 ToGreyScalePS( SCR_VS_OUTPUT In ) : SV_Target
{
    float4 pixColor = SourceImgTexture.Sample( g_samLinear, In.Tex );
    pixColor = 0.3*pixColor.x + 0.59*pixColor.y + 0.11*pixColor.z;
    return pixColor;
}

technique10 ToGreyScale
{
	pass P0
	{
        SetVertexShader( CompileShader( vs_4_0, ToGreyScaleVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0,  ToGreyScalePS() ) );
        
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepthTest, 0 );
	}
}
