#include "DXUT.h"
#include "DXUTgui.h"
#include "DXUTsettingsdlg.h"
#include "DXUTcamera.h"
#include "SDKmisc.h"

#include "FFTDX10.h"
#include "FFTDX10Helper.h"

#include <strstream>
extern CUIState			g_StateChanged;			// Global Sate

extern CD3DSettingsDlg	g_D3DSettingsDlg; 

const double M_PI = 3.1415926535897932384626433832795029;

void GetWeights(int NumPoints, int K, float &Wr, float &Wi)
{
	Wr = float(cosl(2.0*M_PI*K/(float)NumPoints));
	Wi = float(-sinl(2.0*M_PI*K/(float)NumPoints));
}

#pragma warning(disable : 4244)

void BitReverse(float *Indices, int N, int n)
{
	unsigned int mask = 0x1;
	for (int j = 0; j < N; j++)
	{
		unsigned int val = 0x0;
		int temp = int(Indices[j]);
		for (int i = 0; i < n; i++)
		{
			unsigned int t = (mask & temp);
			val = (val << 1) | t;
			temp = temp >> 1;
		}
		Indices[j] = val;
	}
}

void GenerateWeights(int NumPoints, int NumButterflies, float **Wr, float **Wi)
{
	int groups = NumPoints/2;
	int numKs = 1;
	for (int i = 0; i < NumButterflies; i++)
	{
		int start = 0;
		int end = numKs;
		
		for (int b = 0; b < groups; b++)
		{
			for (int k = start, K=0; k < end; k++,K++)
			{
				float wr, wi;
				GetWeights(NumPoints, K*groups, wr, wi);
				Wr[i][k] = wr;
				Wi[i][k] = wi;
				Wr[i][k+numKs] = -wr;
				Wi[i][k+numKs] = -wi;
			}
			start += 2*numKs;
			end = start + numKs;
		}
		groups = groups >> 1;
		numKs = numKs << 1;
	}
}

void GenerateIndices(float **Indices, int NumPoints, int NumButterflies)
{
	int numIters = 1;
	int offset, step;
	offset = NumPoints;
	for (int b = 0; b < NumButterflies; b++)
	{
		offset = offset/2;
		step = 2*offset;
		int p = 0;
		int start = 0;
		int end = step;
		for (int i = 0; i < numIters; i++)
		{
			for (int j = start, k = p,l = 0; j < end; j+=2,l+=2,k++)
			{
				Indices[b][j] = k;
				Indices[b][j+1] = k+offset;
				Indices[b][l+end] = k;
				Indices[b][l+end+1] = k+offset;
			}
			start += 2*step;
			end += 2*step;
			p += step;
		}

		numIters = numIters << 1;
	}
	BitReverse(Indices[8], 2*NumPoints, NumButterflies);
}

void CreateButterflyTexture(CTexture2D& ButterflyTexture, UINT Width, UINT NumButterflies)
{
	D3D10_TEXTURE2D_DESC descSI;
	ButterflyTexture.GetDesc(&descSI);

	assert(Width == descSI.Width);

	float **Indices = new float*[NumButterflies];
	float **Wr = new float*[NumButterflies];
	float **Wi = new float*[NumButterflies];
	
	for (int i = 0; i < (int)NumButterflies; i++)
	{
		Indices[i] = new float[2*Width];
		Wr[i] = new float[Width];
		Wi[i] = new float[Width];
	}
	GenerateIndices(Indices, Width, NumButterflies);
	GenerateWeights(Width, NumButterflies, Wr, Wi);

	// Scrambled indices
	D3D10_MAPPED_TEXTURE2D mappedTexSI;
	ButterflyTexture.m_Texture2D->Map( D3D10CalcSubresource(0, 0, 1), D3D10_MAP_WRITE_DISCARD, 0, &mappedTexSI );
	FLOAT* pTexelsSI = (FLOAT*)mappedTexSI.pData;

	for( UINT row = 0; row < NumButterflies; row++ )
	{
		UINT rowStart = row * mappedTexSI.RowPitch/4;
		for( UINT col = 0; col < descSI.Width; col++ )
		{
			UINT colStart = col * 4;

			pTexelsSI[rowStart + colStart + 0] = (float)(Indices[NumButterflies-row-1][2*col]+0.5)/float(Width); // R
			pTexelsSI[rowStart + colStart + 1] = (float)(Indices[NumButterflies-row-1][2*col+1]+0.5)/float(Width); // G

			pTexelsSI[rowStart + colStart + 2] = Wr[row][col]; // B
			pTexelsSI[rowStart + colStart + 3] = Wi[row][col]; // A
		}
	}

	for (int i = 0; i < (int)NumButterflies; i++)
	{
		delete Indices[i];
		delete Wr[i];
		delete Wi[i];
	}
	delete [] Indices;
	delete [] Wr;
	delete [] Wi;

	ButterflyTexture.m_Texture2D->Unmap( D3D10CalcSubresource(0, 0, 1) );
}

void SetViewport(ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc)
{
	D3D10_VIEWPORT vp;
    vp.Width = pBackBufferSurfaceDesc->Width;
    vp.Height = pBackBufferSurfaceDesc->Height;
    vp.MinDepth = 0.0f;
    vp.MaxDepth = 1.0f;
    vp.TopLeftX = 0;
    vp.TopLeftY = 0;
    pd3dDevice->RSSetViewports( 1, &vp );
}

struct SVertexTexcoords
{
	D3DXVECTOR3 v3;
	D3DXVECTOR2 v2;
};

void FFTDX10::Render(ID3D10EffectTechnique* pTechnique)
{
	ID3D10Device* pd3dDevice = m_D3DDevice;

	pd3dDevice->IASetInputLayout(m_VertexLayout);

	if (m_VertexBuffer)
	{
		UINT Strides;
		UINT Offsets;

		Strides = (UINT)sizeof(SVertexTexcoords);
		Offsets = 0;
		pd3dDevice->IASetVertexBuffers(0, 1, &m_VertexBuffer, &Strides, &Offsets);
		pd3dDevice->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST);

		D3D10_TECHNIQUE_DESC techDesc;
		pTechnique->GetDesc(&techDesc);

		for(UINT p = 0; p < techDesc.Passes; ++p)
		{
			pTechnique->GetPassByIndex(p)->Apply(0);
			pd3dDevice->Draw(6, 0);
		}
	}
}

void FFTDX10::PingPongRenderTargetAndSourceTexture()
{
	if (m_isPingRenderTarget)
		SetPingAsRenderTarget();
	else
		SetPongAsRenderTarget();
}

void FFTDX10::Butterfly(ID3D10EffectTechnique* pTechnique, int numPasses)
{
	ID3D10ShaderResourceView *const pSRV[1] = {NULL};

	for (int pass = 0; pass < numPasses; pass++)
	{
		m_BflyPassNumVar->SetFloat((float)(float(pass)+0.5)/float(numPasses));

		m_D3DDevice->PSSetShaderResources(0, 1, pSRV);
		
		PingPongRenderTargetAndSourceTexture();
		Render(pTechnique);
	}
}

void FFTDX10::Render()
{
	ID3D10Device* pd3dDevice = m_D3DDevice;

	ID3D10RenderTargetView* oldRTViews[1] = { NULL };
	ID3D10DepthStencilView* oldDSView = NULL;
	pd3dDevice->OMGetRenderTargets(1, oldRTViews, &oldDSView);

	// Save the old viewport
	D3D10_VIEWPORT oldViewport;
	UINT numViewports = 1;
	pd3dDevice->RSGetViewports(&numViewports, &oldViewport);

	DXGI_SURFACE_DESC surfaceDesc;
	surfaceDesc.Height = m_Height;
	surfaceDesc.Width = m_Width;

	ID3D10ShaderResourceView *const pSRV[1] = {NULL};
	pd3dDevice->PSSetShaderResources(0, 1, pSRV);

	if (g_StateChanged.computeFFTEachFrame == true)
	{
		// First render the image to m_PongRTTexture texture
		m_PongRTTexture.SetTextureAsRenderTargetAndClear();
		m_SrcImageResource->SetResource(m_SrcImageTexture.GetResourceView());
			
		SetViewport(pd3dDevice, &surfaceDesc);
		Render(m_RenderTechnique); // now we have the grey scale image in PongRTTexture!

		m_BflyValsResource->SetResource(m_ButterflyTexture.GetResourceView());

		if (g_StateChanged.fftType != CUIState::FFT_NONE)
		{
			m_isPingRenderTarget = true; // since we just wrote to Pong
			if (g_StateChanged.fftType != CUIState::FFT_Y)
				Butterfly(m_HButterflyTechnique, m_NumButterfliesX);

			if (g_StateChanged.fftType != CUIState::FFT_X)
				Butterfly(m_VButterflyTechnique, m_NumButterfliesY);

			if (g_StateChanged.fftType == CUIState::IFFT_XY)
			{
				Butterfly(m_HButterflyTechnique, m_NumButterfliesX);
				Butterfly(m_VButterflyTechnique, m_NumButterfliesY);
			}
		}
		else
			m_isPingRenderTarget = true;

		if (m_isPingRenderTarget)
			m_FinalRenderedImage = m_PongRTTexture.GetResourceView();
		else
			m_FinalRenderedImage = m_PingRTTexture.GetResourceView();
	}
	pd3dDevice->PSSetShaderResources(0, 1, pSRV);
	m_SrcImageResource->SetResource(m_FinalRenderedImage);
	// Restore old view port
	pd3dDevice->RSSetViewports(1, &oldViewport);

	// Restore old RT and DS buffer views
	pd3dDevice->OMSetRenderTargets(1, oldRTViews, oldDSView);
		
	static float scaleFactor = (float)(1.0/float(m_Width*m_Height));
	m_pEffect->GetVariableByName("GreyScale")->AsScalar()->SetBool(false);
	m_pEffect->GetVariableByName("ScaleFactor")->AsScalar()->SetFloat(scaleFactor);
	if (g_StateChanged.fftType == CUIState::IFFT_XY)
	{
		Render(m_ScaleTechnique);
	}
	else if (g_StateChanged.fftType == CUIState::FFT_NONE)
	{
		m_pEffect->GetVariableByName("GreyScale")->AsScalar()->SetBool(true);
		Render(m_DisplayTechnique);
	}
	else
	{
		Render(m_DisplayTechnique);
	}

	SAFE_RELEASE(oldRTViews[0]);
	SAFE_RELEASE(oldDSView);
}

int FFTDX10::Run()
{
	DXUTSetCallbackD3D10DeviceAcceptable(   &FFTDX10Helper::IsDeviceAcceptable,    this );
	DXUTSetCallbackD3D10DeviceCreated(      &FFTDX10Helper::OnCreateDevice,        this );
	DXUTSetCallbackD3D10SwapChainResized(   &FFTDX10Helper::OnResizedSwapChain,    this );
	DXUTSetCallbackD3D10SwapChainReleasing( &FFTDX10Helper::OnReleasingSwapChain,  this );
	DXUTSetCallbackD3D10DeviceDestroyed(    &FFTDX10Helper::OnDestroyDevice,       this );
	DXUTSetCallbackD3D10FrameRender(        &FFTDX10Helper::OnFrameRender,         this );
	DXUTSetCallbackMsgProc(                 &FFTDX10Helper::MsgProc,               this );
	DXUTSetCallbackKeyboard(                &FFTDX10Helper::KeyboardProc,          this );
	DXUTSetCallbackFrameMove(               &FFTDX10Helper::OnFrameMove,           this );
	DXUTSetCallbackDeviceChanging(          &FFTDX10Helper::ModifyDeviceSettings,  this );
	DXUTSetCallbackDeviceRemoved(           &FFTDX10Helper::OnDeviceRemoved,       this );
	InitApp();
	DXUTInit( true, true, NULL ); // Parse the command line, show msgboxes on error, no extra command line params
	DXUTSetCursorSettings( true, true ); // Show the cursor and clip it when in full screen
	DXUTCreateWindow( L"FFTDX10" );
	DXUTCreateDevice( true, 512, 512 );
	DXUTMainLoop(); // Enter into the DXUT render loop

	return DXUTGetExitCode();

}

FFTDX10::FFTDX10() : m_FinalRenderedImage(NULL),
					   m_CurrentTechnique(NULL),
					   m_VertexBuffer(NULL),
					   m_VertexLayout(NULL),
					   m_D3DDevice(NULL),
					   m_isPingRenderTarget(false)
{
}

HRESULT FFTDX10::CreateVertexBuffer()
{
	HRESULT hr = S_FALSE;
	static SVertexTexcoords vertices[6] =
    {
		{ D3DXVECTOR3(-1.0f,  1.0f, 0.0f), D3DXVECTOR2(0.0f, 0.0f) },
		{ D3DXVECTOR3( 1.0f, -1.0f, 0.0f), D3DXVECTOR2(1.0f, 1.0f) },
		{ D3DXVECTOR3(-1.0f, -1.0f, 0.0f), D3DXVECTOR2(0.0f, 1.0f) },

		{ D3DXVECTOR3(-1.0f,  1.0f, 0.0f), D3DXVECTOR2(0.0f, 0.0f) },
		{ D3DXVECTOR3( 1.0f,  1.0f, 0.0f), D3DXVECTOR2(1.0f, 0.0f) },
		{ D3DXVECTOR3( 1.0f, -1.0f, 0.0f), D3DXVECTOR2(1.0f, 1.0f) },
    };

	if (!m_VertexBuffer)
	{
		D3D10_BUFFER_DESC bd;
		bd.Usage = D3D10_USAGE_DEFAULT;
		bd.ByteWidth = sizeof(vertices);
		bd.BindFlags = D3D10_BIND_VERTEX_BUFFER;
		bd.CPUAccessFlags = 0;
		bd.MiscFlags = 0;
	    
		D3D10_SUBRESOURCE_DATA InitData;
		ZeroMemory(&InitData, sizeof(InitData));
		InitData.pSysMem = vertices;
	    
		hr = m_D3DDevice->CreateBuffer(&bd, &InitData, &m_VertexBuffer);
	}
	return hr;
}

void FFTDX10::InitApp()
{
	m_UI.Init();
}
//--------------------------------------------------------------------------------------
// Reject any D3D10 devices that aren't acceptable by returning false
//--------------------------------------------------------------------------------------
bool CALLBACK FFTDX10::IsDeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, DXGI_FORMAT BackBufferFormat, bool bWindowed )
{
	// No CAPS for DX10 devices :)
	return true;
}


//--------------------------------------------------------------------------------------
// Called right before creating a D3D9 or D3D10 device, allowing the app to modify the device settings as needed
//--------------------------------------------------------------------------------------
bool CALLBACK FFTDX10::ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings   )
{
	// Turn vsync off
	//	pDeviceSettings->d3d10. pp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
	//	g_SettingsDlg.GetDialogControl()->GetComboBox( DXUTSETTINGSDLG_PRESENT_INTERVAL )->SetSelectedByText(L"Off");

	// For the first device created if its a REF device, optionally display a warning dialog box
	static bool s_bFirstTime = true;
	if( s_bFirstTime )
	{
		s_bFirstTime = false;
		//if( (DXUT_D3D9_DEVICE == pDeviceSettings->ver && pDeviceSettings->d3d9.DeviceType == D3DDEVTYPE_REF) ||
		if( ( pDeviceSettings->ver == DXUT_D3D10_DEVICE) && (pDeviceSettings->d3d10.DriverType == D3D10_DRIVER_TYPE_REFERENCE ))
			DXUTDisplaySwitchingToREFWarning( pDeviceSettings->ver );
	}
	return true;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that aren't dependant on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK FFTDX10::OnCreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc )
{
	HRESULT hr;
	m_D3DDevice = pd3dDevice;

	m_UI.Create(pd3dDevice);   
	m_SrcImageResource	= NULL;
	m_BflyValsResource	= NULL;

	// Read the D3DX effect file doing basic T&L 
	WCHAR str[MAX_PATH];
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"FFT.fx" ) );

	// If this fails, there should be debug output as to where the .fx file failed to compile
	DWORD dwShaderFlags = D3D10_SHADER_ENABLE_STRICTNESS;
#if defined( DEBUG ) || defined( _DEBUG )
	dwShaderFlags |= D3D10_SHADER_DEBUG;

	ID3D10Blob *ppErrors;
	D3DX10CreateEffectFromFile( str, NULL, NULL, "fx_4_0", dwShaderFlags, 0, 
		pd3dDevice, NULL, NULL, &m_pEffect, &ppErrors, NULL);
	
	LPVOID l_pError = NULL;
	if( ppErrors )
	{
		l_pError = ppErrors->GetBufferPointer();
		OutputDebugStringA((char *)l_pError);
	}
#else
	V_RETURN( D3DX10CreateEffectFromFile( str, NULL, NULL, "fx_4_0", dwShaderFlags, 0, 
		pd3dDevice, NULL, NULL, &m_pEffect, NULL, NULL) );
#endif // _DEBUG


	////////////////
	//// Techniques
	////////////////
	m_RenderTechnique     = m_pEffect->GetTechniqueByName( "ToGreyScale" );
	m_DisplayTechnique    = m_pEffect->GetTechniqueByName( "Display" );
	m_HButterflyTechnique = m_pEffect->GetTechniqueByName( "HorizontalButterfly" );
	m_VButterflyTechnique = m_pEffect->GetTechniqueByName( "VerticalButterfly" );
	m_ScaleTechnique	  = m_pEffect->GetTechniqueByName( "Scale" );

	// Create and set the vertex input layout
	D3D10_PASS_DESC PassDesc;
		
	CreateVertexBuffer();
	m_RenderTechnique->GetPassByIndex(0)->GetDesc(&PassDesc);

	// Define the input layout
	const D3D10_INPUT_ELEMENT_DESC default_layout[] =
	{
		{ "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D10_INPUT_PER_VERTEX_DATA, 0 },
		{ "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
	};
	const unsigned int numDefaultElements = sizeof(default_layout)/sizeof(D3D10_INPUT_ELEMENT_DESC);

	V_RETURN(pd3dDevice->CreateInputLayout(default_layout, numDefaultElements, PassDesc.pIAInputSignature, 
			PassDesc.IAInputSignatureSize, &m_VertexLayout));
	pd3dDevice->IASetInputLayout(m_VertexLayout);

	////////////////
	//// Resources
	////////////////
	m_SrcImageResource = m_pEffect->GetVariableByName( "SourceImgTexture" )->AsShaderResource();
	m_BflyValsResource = m_pEffect->GetVariableByName( "ButterflyTexture" )->AsShaderResource();

	////////////////
	//// Variables
	////////////////
	m_BflyPassNumVar	    = m_pEffect->GetVariableByName("ButterflyPassNumber")->AsScalar();

	// Initialize Camera
	// Setup the camera's view parameters
	D3DXVECTOR3 vecEye(0.0f, 0.6f, -200.0f);
	D3DXVECTOR3 vecAt (0.0f, 0.6f, 0.0f);
	m_Camera.SetViewParams( &vecEye, &vecAt );

    V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"Images\\banflwr.jpg" ) );
	
	m_RenderTechnique->GetPassByIndex(0)->GetDesc(&PassDesc);

	////////////////
	//// Textures
	////////////////
	m_SrcImageTexture.LoadFromFile(pd3dDevice, str); 

	D3D10_TEXTURE2D_DESC desc;
	m_SrcImageTexture.GetDesc(&desc);

	m_Width = desc.Width;
	m_Height = desc.Height;
	m_NumButterfliesX = (int)(logf(float(desc.Width))/logf(2.0));
	m_NumButterfliesY = (int)(logf(float(desc.Height))/logf(2.0));

	V_RETURN(m_PingRTTexture.CreateRenderTargetTexture(pd3dDevice, desc.Width, desc.Height, DXGI_FORMAT_R32G32B32A32_FLOAT));
	V_RETURN(m_PongRTTexture.CreateRenderTargetTexture(pd3dDevice, desc.Width, desc.Height, DXGI_FORMAT_R32G32B32A32_FLOAT));

	V_RETURN( m_ButterflyTexture.CreateRWTexture(pd3dDevice, desc.Width, m_NumButterfliesX, DXGI_FORMAT_R32G32B32A32_FLOAT) );
	CreateButterflyTexture(m_ButterflyTexture, desc.Width, m_NumButterfliesX);
	return S_OK;
}

void CALLBACK FFTDX10::OnGUIEvent(UINT nEvent, int nControlID, CDXUTControl *pControl, void* pUserContext  )
{
	m_UI.OnGUIEvent(nEvent, nControlID, pControl, pUserContext);
}

//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------
void CALLBACK FFTDX10::OnFrameMove( double fTime, float fElapsedTime   )
{
	// Update the camera's position based on user input 
	m_Camera.FrameMove( fElapsedTime );
}


//--------------------------------------------------------------------------------------
// Render the scene using the D3D10 device
//--------------------------------------------------------------------------------------
void CALLBACK FFTDX10::OnFrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime   )
{
    // Clear render target and the depth stencil 
    float ClearColor[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
    pd3dDevice->ClearRenderTargetView( DXUTGetD3D10RenderTargetView(), ClearColor );

	Render();
	m_UI.Render(fElapsedTime);
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that depend on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK FFTDX10::OnSwapChainResized( ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc  )
{
	m_UI.OnD3D10SwapChainResized(pd3dDevice,pBackBufferSurfaceDesc);
	// Setup the camera's projection parameters
	float fAspectRatio = pBackBufferSurfaceDesc->Width / (FLOAT)pBackBufferSurfaceDesc->Height;
	m_Camera.SetProjParams( D3DX_PI/4, fAspectRatio, 0.1f, 5000.0f );
	m_Camera.SetWindow( pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height );
	m_Camera.SetButtonMasks( 0, MOUSE_WHEEL, MOUSE_LEFT_BUTTON | MOUSE_MIDDLE_BUTTON | MOUSE_RIGHT_BUTTON );

	return S_OK;
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10ResizedSwapChain 
//--------------------------------------------------------------------------------------
void CALLBACK FFTDX10::OnSwapChainReleasing(  )
{
	m_UI.OnD3D10SwapChainReleasing();
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10CreateDevice 
//--------------------------------------------------------------------------------------
void CALLBACK FFTDX10::OnDestroyDevice(   )
{
	DXUTGetGlobalResourceCache().OnDestroyDevice();
}


//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------
LRESULT CALLBACK FFTDX10::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, 
                          bool* pbNoFurtherProcessing )
{
	m_Camera.HandleMessages( hWnd, uMsg, wParam, lParam );
	m_UI.MsgProc( hWnd, uMsg, wParam, lParam, pbNoFurtherProcessing, NULL );
	return 0;
}


//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------
void CALLBACK FFTDX10::KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown )
{
}

//--------------------------------------------------------------------------------------
// Call if device was removed.  Return true to find a new device, false to quit
//--------------------------------------------------------------------------------------
bool CALLBACK FFTDX10::OnDeviceRemoved(  )
{
    return true;
}
